@@ -1,5 +1,5 @@
 /*
-	Copyright (C) 2003  Sridhar <r_sridhar@users.sf.net>
+	Copyright (C) 2003  Sridhar Ratna <srid@nearfar.org>
 	
 	This file is part of GNUSim8085.
 
@@ -533,12 +533,20 @@ _eef_inst_func_push (eef_addr_t opnd_addr, gchar x)
 {
 	eef_data_t *high, *low;
 
-	eef_loc_of_regpair (x, &high, &low);
-	g_assert (high);
-
-	eef_stack_push_byte (*high);
-	eef_stack_push_byte (*low);
+	if ('P' != x)
+	{
+		eef_loc_of_regpair (x, &high, &low);
+		g_assert (high);
 
+		eef_stack_push_byte (*high);
+		eef_stack_push_byte (*low);	
+	} 
+	else 
+	{		
+		eef_stack_push_byte (sys.reg.a);
+		eef_stack_push_byte (eef_flag_to_data(sys.flag));
+	}
+	
 	/* stack cb */
 	eef_call_stack_cb (TRUE, x);
 
@@ -550,12 +558,21 @@ _eef_inst_func_pop (eef_addr_t opnd_addr, gchar x)
 {
 	eef_data_t *high, *low;
 
-	eef_loc_of_regpair (x, &high, &low);
-	g_assert (high);
+	if ('P' != x)
+	{
+		eef_loc_of_regpair (x, &high, &low);
+		g_assert (high);
 
-	*low = eef_stack_pop_byte ();
-	*high = eef_stack_pop_byte ();
+		*low = eef_stack_pop_byte ();
+		*high = eef_stack_pop_byte ();
 
+	} 
+	else 
+	{
+		eef_data_to_flag(eef_stack_pop_byte (), &sys.flag);
+		sys.reg.a = eef_stack_pop_byte ();
+	}
+	
 	/* stack cb */
 	eef_call_stack_cb (FALSE, x);
 
