@@ -1,5 +1,5 @@
 /*
-	Copyright (C) 2003  Sridhar <r_sridhar@users.sf.net>
+	Copyright (C) 2003  Sridhar Ratna <srid@nearfar.org>
 	
 	This file is part of GNUSim8085.
 
@@ -242,6 +242,7 @@ eef_loc_of_regpair (gchar x, eef_data_t ** h, eef_data_t ** l)
 	case 'P':
 		*h = &sys.reg.pswh;
 		*l = &sys.reg.pswl;
+		g_warning ("Unwanted %c reached", x);
 		break;
 	case 'M':
 		*h = NULL;
@@ -495,3 +496,55 @@ eef_regpair_get_another (gchar x)
 	g_assert_not_reached ();
 	return 0; /* supress warning */
 }
+
+/* Position of flag in flag register 
+        sz0h0p1c
+*/
+#define C_FLAG_POSITION 0x01
+#define AC_FLAG_POSITION 0x10
+#define P_FLAG_POSITION 0x04
+#define Z_FLAG_POSITION 0x40
+#define S_FLAG_POSITION 0x80
+/* Default value of flag register */
+#define DEFAULT_FLAG 0x02
+
+eef_data_t 
+eef_flag_to_data (EefFlag flag)
+{
+	eef_data_t td;
+	td = DEFAULT_FLAG;
+	td |= (0 == flag.c) ? 0 : C_FLAG_POSITION;
+	td |= (0 == flag.ac) ? 0 : AC_FLAG_POSITION;
+	td |= (0 == flag.p) ? 0 : P_FLAG_POSITION;
+	td |= (0 == flag.z) ? 0 : Z_FLAG_POSITION;
+	td |= (0 == flag.s) ? 0 : S_FLAG_POSITION;
+        
+        return td;
+/*return DEFAULT_FLAG | 
+((0 == flag.c) ? 0 : C_FLAG_POSITION) |
+((0 == flag.ac) ? 0 : AC_FLAG_POSITION) | 
+((0 == flag.p) ? 0 : P_FLAG_POSITION) | 
+((0 == flag.z) ? 0 : Z_FLAG_POSITION) | 
+((0 == flag.s) ? 0 : S_FLAG_POSITION);*/
+}
+
+//EefFlag *
+void
+eef_data_to_flag (eef_data_t from, EefFlag *flag)
+{       
+/*	flag.c  = (0 == (from && C_FLAG_POSITION)) ? 0 : 1;
+	flag.ac = (0 == (from && AC_FLAG_POSITION)) ? 0 : 1;
+	flag.p  = (0 == (from && P_FLAG_POSITION)) ? 0 : 1;
+	flag.s  = (0 == (from && S_FLAG_POSITION)) ? 0 : 1;
+	flag.z  = (0 == (from && Z_FLAG_POSITION)) ? 0 : 1;*/
+
+//	EefFlag * flag;
+	
+	flag->c  = (0 == (from & C_FLAG_POSITION)) ? 0 : 1;
+	flag->ac = (0 == (from & AC_FLAG_POSITION)) ? 0 : 1;
+	flag->p  = (0 == (from & P_FLAG_POSITION)) ? 0 : 1;
+	flag->s  = (0 == (from & S_FLAG_POSITION)) ? 0 : 1;
+	flag->z  = (0 == (from & Z_FLAG_POSITION)) ? 0 : 1; 
+	
+//	return flag;
+}
