@@ -1,6 +1,6 @@
 // Scintilla source code edit control
 // PlatGTK.cxx - implementation of platform facilities on GTK+/Linux
-// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
+// Copyright 1998-2004 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 #include <string.h>
@@ -37,8 +37,7 @@
 
 #if GTK_MAJOR_VERSION >= 2
 #define USE_PANGO 1
-#include <iconv.h>
-const iconv_t iconvhBad = (iconv_t)(-1);
+#include "Converter.h"
 #endif
 
 #ifdef _MSC_VER
@@ -270,7 +269,7 @@ static const char *CharacterSetName(int characterSet) {
 	case SC_CHARSET_DEFAULT:
 		return "iso8859-*";
 	case SC_CHARSET_BALTIC:
-		return "*-*";
+		return "iso8859-13";
 	case SC_CHARSET_CHINESEBIG5:
 		return "*-*";
 	case SC_CHARSET_EASTEUROPE:
@@ -287,6 +286,8 @@ static const char *CharacterSetName(int characterSet) {
 		return "*-*";
 	case SC_CHARSET_RUSSIAN:
 		return "*-r";
+	case SC_CHARSET_CYRILLIC:
+		return "*-cp1251";
 	case SC_CHARSET_SHIFTJIS:
 		return "jisx0208.1983-*";
 	case SC_CHARSET_SYMBOL:
@@ -303,6 +304,8 @@ static const char *CharacterSetName(int characterSet) {
 		return "*-*";
 	case SC_CHARSET_THAI:
 		return "*-1";
+	case SC_CHARSET_8859_15:
+		return "iso8859-15";
 	default:
 		return "*-*";
 	}
@@ -646,7 +649,8 @@ Font::Font() : id(0) {}
 
 Font::~Font() {}
 
-void Font::Create(const char *faceName, int characterSet, int size, bool bold, bool italic) {
+void Font::Create(const char *faceName, int characterSet, int size,
+	bool bold, bool italic, bool) {
 	Release();
 	id = FontCached::FindOrCreate(faceName, characterSet, size, bold, italic);
 }
@@ -669,9 +673,9 @@ class SurfaceImpl : public Surface {
 #ifdef USE_PANGO
 	PangoContext *pcontext;
 	PangoLayout *layout;
-	iconv_t iconvh;
+	Converter conv;
 	int characterSet;
-	void SetIconv(int characterSet_);
+	void SetConverter(int characterSet_);
 #endif
 public:
 	SurfaceImpl();
@@ -718,8 +722,6 @@ public:
 	void SetDBCSMode(int codePage);
 };
 
-#ifdef USE_PANGO
-
 const char *CharacterSetID(int characterSet) {
 	switch (characterSet) {
 	case SC_CHARSET_ANSI:
@@ -727,15 +729,15 @@ const char *CharacterSetID(int characterSet) {
 	case SC_CHARSET_DEFAULT:
 		return "LATIN1";
 	case SC_CHARSET_BALTIC:
-		return "BALTIC";
+		return "ISO-8859-13";
 	case SC_CHARSET_CHINESEBIG5:
 		return "BIG-5";
 	case SC_CHARSET_EASTEUROPE:
-		return "ISO8859-2";
+		return "ISO-8859-2";
 	case SC_CHARSET_GB2312:
 		return "GB2312";
 	case SC_CHARSET_GREEK:
-		return "ISO8859-7";
+		return "ISO-8859-7";
 	case SC_CHARSET_HANGUL:
 		return "";
 	case SC_CHARSET_MAC:
@@ -744,37 +746,37 @@ const char *CharacterSetID(int characterSet) {
 		return "ASCII";
 	case SC_CHARSET_RUSSIAN:
 		return "KOI8-R";
+	case SC_CHARSET_CYRILLIC:
+		return "CP1251";
 	case SC_CHARSET_SHIFTJIS:
 		return "SHIFT-JIS";
 	case SC_CHARSET_SYMBOL:
 		return "";
 	case SC_CHARSET_TURKISH:
-		return "ISO8859-9";
+		return "ISO-8859-9";
 	case SC_CHARSET_JOHAB:
 		return "JOHAB";
 	case SC_CHARSET_HEBREW:
-		return "ISO8859-8";
+		return "ISO-8859-8";
 	case SC_CHARSET_ARABIC:
-		return "ISO8859-6";
+		return "ISO-8859-6";
 	case SC_CHARSET_VIETNAMESE:
 		return "";
 	case SC_CHARSET_THAI:
-		return "ISO8859-1";
+		return "ISO-8859-1";
+	case SC_CHARSET_8859_15:
+		return "ISO-8859-15";
 	default:
 		return "";
 	}
 }
 
-void SurfaceImpl::SetIconv(int characterSet_) {
+#ifdef USE_PANGO
+
+void SurfaceImpl::SetConverter(int characterSet_) {
 	if (characterSet != characterSet_) {
-		if (iconvh != iconvhBad)
-			iconv_close(iconvh);
-		iconvh = iconvhBad;
 		characterSet = characterSet_;
-		const char *source = CharacterSetID(characterSet);
-		if (*source) {
-			iconvh = iconv_open("UTF8", source);
-		}
+		conv.Open("UTF-8", CharacterSetID(characterSet));
 	}
 }
 #endif
@@ -782,7 +784,7 @@ void SurfaceImpl::SetIconv(int characterSet_) {
 SurfaceImpl::SurfaceImpl() : et(singleByte), drawable(0), gc(0), ppixmap(0),
 x(0), y(0), inited(false), createdGC(false)
 #ifdef USE_PANGO
-, pcontext(0), layout(0), iconvh(iconvhBad), characterSet(-1)
+, pcontext(0), layout(0), characterSet(-1)
 #endif
 {
 }
@@ -809,9 +811,7 @@ void SurfaceImpl::Release() {
 	if (pcontext)
 		g_object_unref(pcontext);
 	pcontext = 0;
-	if (iconvh != iconvhBad)
-		iconv_close(iconvh);
-	iconvh = iconvhBad;
+	conv.Close();
 	characterSet = -1;
 #endif
 	x = 0;
@@ -854,6 +854,8 @@ void SurfaceImpl::Init(SurfaceID sid, WindowID WID_NAME) {
 #endif
 	drawable = drawable_;
 	gc = gdk_gc_new(drawable_);
+	// Ask for lines that do not paint the last pixel so is like Win32
+	gdk_gc_set_line_attributes(gc, 0, GDK_LINE_SOLID, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
 	createdGC = true;
 	inited = true;
 }
@@ -874,6 +876,8 @@ void SurfaceImpl::InitPixMap(int width, int height, Surface *surface_, WindowID
 		ppixmap = gdk_pixmap_new(surfImpl->drawable, width, height, -1);
 	drawable = ppixmap;
 	gc = gdk_gc_new(surfImpl->drawable);
+	// Ask for lines that do not paint the last pixel so is like Win32
+	gdk_gc_set_line_attributes(gc, 0, GDK_LINE_SOLID, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
 	createdGC = true;
 	inited = true;
 }
@@ -1030,25 +1034,7 @@ static size_t UTF8Len(char ch) {
 		return 3;
 }
 
-#ifdef USE_PANGO
-static char *UTF8FromIconv(iconv_t iconvh, const char *s, int len) {
-	if (iconvh != ((iconv_t)(-1))) {
-		char *utfForm = new char[len*3+1];
-		char *pin = const_cast<char *>(s);
-		size_t inLeft = len;
-		char *pout = utfForm;
-		size_t outLeft = len*3+1;
-		size_t conversions = iconv(iconvh, &pin, &inLeft, &pout, &outLeft);
-		if (conversions != ((size_t)(-1))) {
-			*pout = '\0';
-			return utfForm;
-		}
-		delete []utfForm;
-	}
-	return 0;
-}
-
-static char *UTF8FromLatin1(const char *s, int len) {
+char *UTF8FromLatin1(const char *s, int &len) {
 	char *utfForm = new char[len*2+1];
 	size_t lenU = 0;
 	for (int i=0;i<len;i++) {
@@ -1061,9 +1047,46 @@ static char *UTF8FromLatin1(const char *s, int len) {
 		}
 	}
 	utfForm[lenU] = '\0';
+	len = lenU;
 	return utfForm;
 }
 
+#ifdef USE_PANGO
+static char *UTF8FromIconv(const Converter &conv, const char *s, int &len) {
+	if (conv) {
+		char *utfForm = new char[len*3+1];
+		char *pin = const_cast<char *>(s);
+		size_t inLeft = len;
+		char *pout = utfForm;
+		size_t outLeft = len*3+1;
+		size_t conversions = conv.Convert(&pin, &inLeft, &pout, &outLeft);
+		if (conversions != ((size_t)(-1))) {
+			*pout = '\0';
+			len = pout - utfForm;
+			return utfForm;
+		}
+		delete []utfForm;
+	}
+	return 0;
+}
+
+// Work out how many bytes are in a character by trying to convert using iconv,
+// returning the first length that succeeds.
+static size_t MultiByteLenFromIconv(const Converter &conv, const char *s, size_t len) {
+	for (size_t lenMB=1; (lenMB<4) && (lenMB <= len); lenMB++) {
+		char wcForm[2];
+		char *pin = const_cast<char *>(s);
+		size_t inLeft = lenMB;
+		char *pout = wcForm;
+		size_t outLeft = 2;
+		size_t conversions = conv.Convert(&pin, &inLeft, &pout, &outLeft);
+		if (conversions != ((size_t)(-1))) {
+			return lenMB;
+		}
+	}
+	return 1;
+}
+
 static char *UTF8FromGdkWChar(GdkWChar *wctext, int wclen) {
 	char *utfForm = new char[wclen*3+1];	// Maximum of 3 UTF-8 bytes per character
 	size_t lenU = 0;
@@ -1084,7 +1107,7 @@ static char *UTF8FromGdkWChar(GdkWChar *wctext, int wclen) {
 	return utfForm;
 }
 
-static char *UTF8FromDBCS(const char *s, int len) {
+static char *UTF8FromDBCS(const char *s, int &len) {
 	GdkWChar *wctext = new GdkWChar[len + 1];
 	GdkWChar *wcp = wctext;
 	int wclen = gdk_mbstowcs(wcp, s, len);
@@ -1097,9 +1120,22 @@ static char *UTF8FromDBCS(const char *s, int len) {
 
 	char *utfForm = UTF8FromGdkWChar(wctext, wclen);
 	delete []wctext;
+	len = strlen(utfForm);
 	return utfForm;
 }
 
+static size_t UTF8CharLength(const char *s) {
+	const unsigned char *us = reinterpret_cast<const unsigned char *>(s);
+	unsigned char ch = *us;
+	if (ch < 0x80) {
+		return 1;
+	} else if (ch < 0x80 + 0x40 + 0x20) {
+		return 2;
+	} else {
+		return 3;
+	}
+}
+
 #endif
 
 // On GTK+, wchar_t is 4 bytes
@@ -1118,23 +1154,20 @@ void SurfaceImpl::DrawTextBase(PRectangle rc, Font &font_, int ybase, const char
 			if (et == UTF8) {
 				pango_layout_set_text(layout, s, len);
 			} else {
-				if (et == dbcs) {
-					// Convert to utf8
-					utfForm = UTF8FromDBCS(s, len);
+				if (!utfForm) {
+					SetConverter(PFont(font_)->characterSet);
+					utfForm = UTF8FromIconv(conv, s, len);
 				}
-				if (!utfForm) {	// DBCS failed so treat as iconv
-					SetIconv(PFont(font_)->characterSet);
-					utfForm = UTF8FromIconv(iconvh, s, len);
+				if (!utfForm) {	// iconv failed so try DBCS if DBCS mode
+					if (et == dbcs) {
+						// Convert to utf8
+						utfForm = UTF8FromDBCS(s, len);
+					}
 				}
-				//~ if (!utfForm) {	// DBCS failed so treat as locale
-					//~ gsize w; // stub
-					//~ utfForm = g_locale_to_utf8(s, len, NULL, &w, NULL);
-					//~ useGFree = static_cast<bool>(utfForm);
-				//~ };
-				if (!utfForm) {	// g_locale_to_utf8 failed so treat as Latin1
+				if (!utfForm) {	// iconv and DBCS failed so treat as Latin1
 					utfForm = UTF8FromLatin1(s, len);
 				}
-				pango_layout_set_text(layout, utfForm, strlen(utfForm));
+				pango_layout_set_text(layout, utfForm, len);
 			}
 			pango_layout_set_font_description(layout, PFont(font_)->pfd);
 			PangoLayoutLine *pll = pango_layout_get_line(layout,0);
@@ -1152,13 +1185,18 @@ void SurfaceImpl::DrawTextBase(PRectangle rc, Font &font_, int ybase, const char
 		bool draw8bit = true;
 		if (et != singleByte) {
 			GdkWChar wctext[maxLengthTextRun];
+			if (len >= maxLengthTextRun)
+				len = maxLengthTextRun-1;
 			int wclen;
 			if (et == UTF8) {
 				wclen = UCS2FromUTF8(s, len,
 					reinterpret_cast<wchar_t *>(wctext), maxLengthTextRun - 1);
 			} else {	// dbcs, so convert using current locale
+				char sMeasure[maxLengthTextRun];
+				memcpy(sMeasure, s, len);
+				sMeasure[len] = '\0';
 				wclen = gdk_mbstowcs(
-					wctext, s, maxLengthTextRun - 1);
+					wctext, sMeasure, maxLengthTextRun - 1);
 			}
 			if (wclen > 0) {
 				draw8bit = false;
@@ -1169,11 +1207,11 @@ void SurfaceImpl::DrawTextBase(PRectangle rc, Font &font_, int ybase, const char
 					gdk_draw_text_wc(drawable, PFont(font_)->pfont, gc,
 							 x, ybase, wcp, lenDraw);
 					wclen -= lenDraw;
-					wcp += lenDraw;
 					if (wclen > 0) {	// Avoid next calculation if possible as may be expensive
 						x += gdk_text_width_wc(PFont(font_)->pfont,
 								       wcp, lenDraw);
 					}
+					wcp += lenDraw;
 				}
 			}
 		}
@@ -1183,10 +1221,10 @@ void SurfaceImpl::DrawTextBase(PRectangle rc, Font &font_, int ybase, const char
 				gdk_draw_text(drawable, PFont(font_)->pfont, gc,
 				              x, ybase, s, lenDraw);
 				len -= lenDraw;
-				s += lenDraw;
 				if (len > 0) {	// Avoid next calculation if possible as may be expensive
 					x += gdk_text_width(PFont(font_)->pfont, s, lenDraw);
 				}
+				s += lenDraw;
 			}
 		}
 	}
@@ -1220,6 +1258,7 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 	if (font_.GetID()) {
 		int totalWidth = 0;
 #ifdef USE_PANGO
+		const int lenPositions = len;
 		if (PFont(font_)->pfd) {
 			if (len == 1) {
 				int width = PFont(font_)->CharWidth(*s, et);
@@ -1233,63 +1272,82 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 			if (et == UTF8) {
 				// Simple and direct as UTF-8 is native Pango encoding
 				pango_layout_set_text(layout, s, len);
+				PangoLayoutIter *iter = pango_layout_get_iter(layout);
+				pango_layout_iter_get_cluster_extents(iter, NULL, &pos);
 				int i = 0;
-				while (i < len) {
-					pango_layout_index_to_pos(layout, i+1, &pos);
-					positions[i++] = PANGO_PIXELS(pos.x);
+				while (pango_layout_iter_next_cluster(iter)) {
+					pango_layout_iter_get_cluster_extents(iter, NULL, &pos);
+					int position = PANGO_PIXELS(pos.x);
+					int curIndex = pango_layout_iter_get_index(iter);
+					while (i < curIndex) {
+						positions[i++] = position;
+					}
 				}
+				while (i < lenPositions)
+					positions[i++] = PANGO_PIXELS(pos.x + pos.width);
+				pango_layout_iter_free(iter);
+				PLATFORM_ASSERT(i == lenPositions);
 			} else {
-				int wclen = 0;
+				int positionsCalculated = 0;
 				if (et == dbcs) {
-					GdkWChar *wctext = new GdkWChar[len + 1];
-					GdkWChar *wcp = wctext;
-					wclen = gdk_mbstowcs(wcp, s, len);
-					if (wclen >= 1 ) {
+					SetConverter(PFont(font_)->characterSet);
+					char *utfForm = UTF8FromIconv(conv, s, len);
+					if (utfForm) {
 						// Convert to UTF-8 so can ask Pango for widths, then
 						// Loop through UTF-8 and DBCS forms, taking account of different
 						// character byte lengths.
-						char *utfForm = UTF8FromGdkWChar(wctext, wclen);
+						Converter convMeasure("UCS-2", CharacterSetID(characterSet));
 						pango_layout_set_text(layout, utfForm, strlen(utfForm));
 						int i = 0;
-						int iU = 0;
-						while (i < len) {
-							iU += UTF8Len(utfForm[iU]);
-							pango_layout_index_to_pos(layout, iU, &pos);
-							size_t lenChar = mblen(s+i, MB_CUR_MAX);
-							while (lenChar--) {
-								positions[i++] = PANGO_PIXELS(pos.x);
+						int utfIndex = 0;
+						PangoLayoutIter *iter = pango_layout_get_iter(layout);
+						pango_layout_iter_get_cluster_extents(iter, NULL, &pos);
+						while (pango_layout_iter_next_cluster(iter)) {
+							pango_layout_iter_get_cluster_extents (iter, NULL, &pos);
+							int position = PANGO_PIXELS(pos.x);
+							int utfIndexNext = pango_layout_iter_get_index(iter);
+							while (utfIndex < utfIndexNext) {
+								size_t lenChar = MultiByteLenFromIconv(convMeasure, s+i, len-i);
+								//size_t lenChar = mblen(s+i, MB_CUR_MAX);
+								while (lenChar--) {
+									positions[i++] = position;
+									positionsCalculated++;
+								}
+								utfIndex += UTF8CharLength(utfForm+utfIndex);
 							}
 						}
+						while (i < lenPositions)
+							positions[i++] = PANGO_PIXELS(pos.x + pos.width);
+						pango_layout_iter_free(iter);
 						delete []utfForm;
+						PLATFORM_ASSERT(i == lenPositions);
 					}
-					delete []wctext;
 				}
-				if (wclen < 1 ) {
+				if (positionsCalculated < 1 ) {
 					// Either Latin1 or DBCS conversion failed so treat as Latin1.
 					bool useGFree = false;
-					SetIconv(PFont(font_)->characterSet);
-					char *utfForm = UTF8FromIconv(iconvh, s, len);
-					//~ if (!utfForm) {	// iconv failed so treat as locale
-						//~ gsize w; // stub
-						//~ utfForm = g_locale_to_utf8(s, len, NULL, &w, NULL);
-						//~ useGFree = static_cast<bool>(utfForm);
-					//~ }
+					SetConverter(PFont(font_)->characterSet);
+					char *utfForm = UTF8FromIconv(conv, s, len);
 					if (!utfForm) {
 						utfForm = UTF8FromLatin1(s, len);
 					}
-					pango_layout_set_text(layout, utfForm, strlen(utfForm));
+					pango_layout_set_text(layout, utfForm, len);
 					int i = 0;
-					int iU = 0;
-					while (i < len) {
-						iU += UTF8Len(utfForm[iU]);
-						pango_layout_index_to_pos(layout, iU, &pos);
+					PangoLayoutIter *iter = pango_layout_get_iter(layout);
+					pango_layout_iter_get_cluster_extents(iter, NULL, &pos);
+					while (pango_layout_iter_next_cluster(iter)) {
+						pango_layout_iter_get_cluster_extents(iter, NULL, &pos);
 						positions[i++] = PANGO_PIXELS(pos.x);
 					}
+					while (i < lenPositions)
+						positions[i++] = PANGO_PIXELS(pos.x + pos.width);
+					pango_layout_iter_free(iter);
 					if (useGFree) {
 						g_free(utfForm);
 					} else {
 						delete []utfForm;
 					}
+					PLATFORM_ASSERT(i == lenPositions);
 				}
 			}
 			if (len == 1) {
@@ -1302,13 +1360,18 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 		bool measure8bit = true;
 		if (et != singleByte) {
 			GdkWChar wctext[maxLengthTextRun];
+			if (len >= maxLengthTextRun)
+				len = maxLengthTextRun-1;
 			int wclen;
 			if (et == UTF8) {
 				wclen = UCS2FromUTF8(s, len,
 					reinterpret_cast<wchar_t *>(wctext), maxLengthTextRun - 1);
 			} else {	// dbcsMode, so convert using current locale
+				char sDraw[maxLengthTextRun];
+				memcpy(sDraw, s, len);
+				sDraw[len] = '\0';
 				wclen = gdk_mbstowcs(
-					wctext, s, maxLengthTextRun - 1);
+					wctext, sDraw, maxLengthTextRun - 1);
 			}
 			if (wclen > 0) {
 				measure8bit = false;
@@ -1318,11 +1381,13 @@ void SurfaceImpl::MeasureWidths(Font &font_, const char *s, int len, int *positi
 				for (int iU = 0; iU < wclen; iU++) {
 					int width = gdk_char_width_wc(gf, wctext[iU]);
 					totalWidth += width;
-					size_t lenChar;
+					int lenChar;
 					if (et == UTF8) {
 						lenChar = UTF8Len(s[i]);
 					} else {
 						lenChar = mblen(s+i, MB_CUR_MAX);
+						if (lenChar < 0)
+							lenChar = 1;
 					}
 					while (lenChar--) {
 						positions[i++] = totalWidth;
@@ -1365,28 +1430,22 @@ int SurfaceImpl::WidthText(Font &font_, const char *s, int len) {
 					utfForm = UTF8FromDBCS(s, len);
 				}
 				if (!utfForm) {	// DBCS failed so treat as iconv
-					SetIconv(PFont(font_)->characterSet);
-					utfForm = UTF8FromIconv(iconvh, s, len);
+					SetConverter(PFont(font_)->characterSet);
+					utfForm = UTF8FromIconv(conv, s, len);
 				}
-				//~ if (!utfForm) {	// iconv failed so treat as locale
-					//~ gsize w;
-					//~ utfForm = g_locale_to_utf8(s, len, NULL, &w, NULL);
-					//~ useGFree = static_cast<bool>(utfForm);
-				//~ };
 				if (!utfForm) {	// g_locale_to_utf8 failed so treat as Latin1
 					utfForm = UTF8FromLatin1(s, len);
 				}
-				pango_layout_set_text(layout, utfForm, strlen(utfForm));
-				len = strlen(utfForm);
+				pango_layout_set_text(layout, utfForm, len);
 			}
-			pango_layout_index_to_pos(layout, len, &pos);
-			int width = PANGO_PIXELS(pos.x);
+			PangoLayoutLine *pangoLine = pango_layout_get_line(layout, 0);
+			pango_layout_line_get_extents(pangoLine, NULL, &pos);
 			if (useGFree) {
 				g_free(utfForm);
 			} else {
 				delete []utfForm;
 			}
-			return width;
+			return PANGO_PIXELS(pos.width);
 		}
 #endif
 		if (et == UTF8) {
@@ -1425,10 +1484,12 @@ int SurfaceImpl::WidthChar(Font &font_, char ch) {
 // rarely used in code.
 
 // This string contains a good range of characters to test for size.
-const char largeSizeString[] = " `~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890"
-                               "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+//const char largeSizeString[] = " `~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890"
+//                               "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+#ifndef FAST_WAY
 const char sizeString[] = "`~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890"
                           "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+#endif
 
 int SurfaceImpl::Ascent(Font &font_) {
 	if (!(font_.GetID()))
@@ -1446,9 +1507,12 @@ int SurfaceImpl::Ascent(Font &font_) {
 		ascent = PFont(font_)->ascent;
 	}
 #endif
-	if (ascent == 0) {
+	if ((ascent == 0) && (PFont(font_)->pfont)) {
 		ascent = PFont(font_)->pfont->ascent;
 	}
+	if (ascent == 0) {
+		ascent = 1;
+	}
 	FontMutexUnlock();
 	return ascent;
 #else
@@ -1529,7 +1593,7 @@ void SurfaceImpl::SetUnicodeMode(bool unicodeMode_) {
 }
 
 void SurfaceImpl::SetDBCSMode(int codePage) {
-	if (codePage == SC_CP_DBCS)
+	if (codePage && (codePage != SC_CP_UTF8))
 		et = dbcs;
 }
 
@@ -1589,6 +1653,19 @@ void Window::SetPositionRelative(PRectangle rc, Window relativeTo) {
 	oy += rc.top;
 	if (oy < 0)
 		oy = 0;
+
+	/* do some corrections to fit into screen */
+	int sizex = rc.right - rc.left;
+	int sizey = rc.bottom - rc.top;
+	int screenWidth = gdk_screen_width();
+	int screenHeight = gdk_screen_height();
+	if (sizex > screenWidth)
+		ox = 0; /* the best we can do */
+	else if (ox + sizex > screenWidth)
+		ox = screenWidth - sizex;
+	if (oy + sizey > screenHeight)
+		oy = screenHeight - sizey;
+
 	gtk_widget_set_uposition(PWidget(id), ox, oy);
 #if 0
 
@@ -1599,7 +1676,7 @@ void Window::SetPositionRelative(PRectangle rc, Window relativeTo) {
 	alloc.height = rc.bottom - rc.top;
 	gtk_widget_size_allocate(id, &alloc);
 #endif
-	gtk_widget_set_usize(PWidget(id), rc.right - rc.left, rc.bottom - rc.top);
+	gtk_widget_set_usize(PWidget(id), sizex, sizey);
 }
 
 PRectangle Window::GetClientPosition() {
@@ -1627,7 +1704,7 @@ void Window::InvalidateRectangle(PRectangle rc) {
 }
 
 void Window::SetFont(Font &) {
-	// TODO
+	// Can not be done generically but only needed for ListBox
 }
 
 void Window::SetCursor(Cursor curs) {
@@ -1663,7 +1740,8 @@ void Window::SetCursor(Cursor curs) {
 		break;
 	}
 
-	gdk_window_set_cursor(PWidget(id)->window, gdkCurs);
+	if (PWidget(id)->window)
+		gdk_window_set_cursor(PWidget(id)->window, gdkCurs);
 	gdk_cursor_destroy(gdkCurs);
 }
 
@@ -1673,16 +1751,25 @@ void Window::SetTitle(const char *s) {
 
 struct ListImage {
 	const char *xpm_data;
+#if GTK_MAJOR_VERSION < 2
 	GdkPixmap *pixmap;
 	GdkBitmap *bitmap;
+#else
+	GdkPixbuf *pixbuf;
+#endif
 };
 
 static void list_image_free(gpointer, gpointer value, gpointer) {
 	ListImage *list_image = (ListImage *) value;
+#if GTK_MAJOR_VERSION < 2
 	if (list_image->pixmap)
 		gdk_pixmap_unref(list_image->pixmap);
 	if (list_image->bitmap)
 		gdk_bitmap_unref(list_image->bitmap);
+#else
+	if (list_image->pixbuf)
+		gdk_pixbuf_unref (list_image->pixbuf);
+#endif
 	g_free(list_image);
 }
 
@@ -1692,10 +1779,20 @@ ListBox::ListBox() {
 ListBox::~ListBox() {
 }
 
+#if GTK_MAJOR_VERSION >= 2
+enum {
+	PIXBUF_COLUMN,
+	TEXT_COLUMN,
+	N_COLUMNS
+};
+#endif
+
 class ListBoxX : public ListBox {
 	WindowID list;
 	WindowID scroller;
+#if GTK_MAJOR_VERSION < 2
 	int current;
+#endif
 	void *pixhash;
 	int lineHeight;
 	XPMSet xset;
@@ -1707,8 +1804,11 @@ public:
 	CallBackAction doubleClickAction;
 	void *doubleClickActionData;
 
-	ListBoxX() : list(0), current(0), pixhash(NULL), desiredVisibleRows(5), maxItemCharacters(0),
+	ListBoxX() : list(0), pixhash(NULL), desiredVisibleRows(5), maxItemCharacters(0),
 		doubleClickAction(NULL), doubleClickActionData(NULL) {
+#if GTK_MAJOR_VERSION < 2
+			current = 0;
+#endif
 	}
 	virtual ~ListBoxX() {
 		if (pixhash) {
@@ -1717,9 +1817,10 @@ public:
 		}
 	}
 	virtual void SetFont(Font &font);
-	virtual void Create(Window &parent, int ctrlID, int lineHeight_, bool unicodeMode_);
+	virtual void Create(Window &parent, int ctrlID, Point location_, int lineHeight_, bool unicodeMode_);
 	virtual void SetAverageCharWidth(int width);
 	virtual void SetVisibleRows(int rows);
+	virtual int GetVisibleRows() const;
 	virtual PRectangle GetDesiredRect();
 	virtual int CaretFromEdge();
 	virtual void Clear();
@@ -1729,13 +1830,13 @@ public:
 	virtual int GetSelection();
 	virtual int Find(const char *prefix);
 	virtual void GetValue(int n, char *value, int len);
-	virtual void Sort();
 	virtual void RegisterImage(int type, const char *xpm_data);
 	virtual void ClearRegisteredImages();
 	virtual void SetDoubleClickAction(CallBackAction action, void *data) {
 		doubleClickAction = action;
 		doubleClickActionData = data;
 	}
+	virtual void SetList(const char* list, char separator, char typesep);
 };
 
 ListBox *ListBox::Allocate() {
@@ -1743,11 +1844,18 @@ ListBox *ListBox::Allocate() {
 	return lb;
 }
 
+#if GTK_MAJOR_VERSION < 2
+static void UnselectionAC(GtkWidget *, gint, gint,
+                        GdkEventButton *, gpointer p) {
+	int *pi = reinterpret_cast<int *>(p);
+	*pi = -1;
+}
 static void SelectionAC(GtkWidget *, gint row, gint,
                         GdkEventButton *, gpointer p) {
 	int *pi = reinterpret_cast<int *>(p);
 	*pi = row;
 }
+#endif
 
 static gboolean ButtonPress(GtkWidget *, GdkEventButton* ev, gpointer p) {
 	ListBoxX* lb = reinterpret_cast<ListBoxX*>(p);
@@ -1759,11 +1867,11 @@ static gboolean ButtonPress(GtkWidget *, GdkEventButton* ev, gpointer p) {
 	return FALSE;
 }
 
-void ListBoxX::Create(Window &, int, int, bool) {
+void ListBoxX::Create(Window &, int, Point, int, bool) {
 	id = gtk_window_new(GTK_WINDOW_POPUP);
 
 	GtkWidget *frame = gtk_frame_new(NULL);
-	gtk_widget_show (frame);
+	gtk_widget_show(frame);
 	gtk_container_add(GTK_CONTAINER(GetID()), frame);
 	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
 	gtk_container_set_border_width(GTK_CONTAINER(frame), 0);
@@ -1775,17 +1883,53 @@ void ListBoxX::Create(Window &, int, int, bool) {
 	gtk_container_add(GTK_CONTAINER(frame), PWidget(scroller));
 	gtk_widget_show(PWidget(scroller));
 
+#if GTK_MAJOR_VERSION < 2
 	list = gtk_clist_new(1);
 	gtk_widget_show(PWidget(list));
 	gtk_container_add(GTK_CONTAINER(PWidget(scroller)), PWidget(list));
 	gtk_clist_set_column_auto_resize(GTK_CLIST(PWidget(list)), 0, TRUE);
 	gtk_clist_set_selection_mode(GTK_CLIST(PWidget(list)), GTK_SELECTION_BROWSE);
+	gtk_signal_connect(GTK_OBJECT(PWidget(list)), "unselect_row",
+	                   GTK_SIGNAL_FUNC(UnselectionAC), &current);
 	gtk_signal_connect(GTK_OBJECT(PWidget(list)), "select_row",
 	                   GTK_SIGNAL_FUNC(SelectionAC), &current);
 	gtk_signal_connect(GTK_OBJECT(PWidget(list)), "button_press_event",
 	                   GTK_SIGNAL_FUNC(ButtonPress), this);
 	gtk_clist_set_shadow_type(GTK_CLIST(PWidget(list)), GTK_SHADOW_NONE);
+#else
+	/* Tree and its model */
+	GtkListStore *store =
+		gtk_list_store_new(N_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING);
 
+	list = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
+	GtkTreeSelection *selection =
+		gtk_tree_view_get_selection(GTK_TREE_VIEW(list));
+	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
+	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(list), FALSE);
+	gtk_tree_view_set_reorderable(GTK_TREE_VIEW(list), FALSE);
+
+	/* Columns */
+	GtkTreeViewColumn *column = gtk_tree_view_column_new();
+	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
+	gtk_tree_view_column_set_title(column, "Autocomplete");
+
+	GtkCellRenderer *renderer = gtk_cell_renderer_pixbuf_new();
+	gtk_tree_view_column_pack_start(column, renderer, FALSE);
+	gtk_tree_view_column_add_attribute(column, renderer,
+										"pixbuf", PIXBUF_COLUMN);
+
+	renderer = gtk_cell_renderer_text_new();
+	gtk_tree_view_column_pack_start(column, renderer, TRUE);
+	gtk_tree_view_column_add_attribute(column, renderer,
+										"text", TEXT_COLUMN);
+
+	gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);
+	gtk_container_add(GTK_CONTAINER(PWidget(scroller)), PWidget(list));
+	gtk_widget_show(PWidget(list));
+
+	gtk_signal_connect(GTK_OBJECT(PWidget(list)), "button_press_event",
+	                   GTK_SIGNAL_FUNC(ButtonPress), this);
+#endif
 	gtk_widget_realize(PWidget(id));
 }
 
@@ -1801,15 +1945,10 @@ void ListBoxX::SetFont(Font &scint_font) {
 		gtk_style_unref(style);
 	}
 #else
-	GtkStyle *styleCurrent = gtk_widget_get_style(GTK_WIDGET(PWidget(list)));
-	GdkFont *fontCurrent = gtk_style_get_font(styleCurrent);
-	if (PFont(scint_font)->pfont) {
-		if (!gdk_font_equal(fontCurrent, PFont(scint_font)->pfont)) {
-			GtkStyle *styleNew = gtk_style_copy(styleCurrent);
-			gtk_style_set_font(styleNew, PFont(scint_font)->pfont);
-			gtk_widget_set_style(GTK_WIDGET(PWidget(list)), styleNew);
-			gtk_style_unref(styleCurrent);
-		}
+	// Only do for Pango font as there have been crashes for GDK fonts
+	if (PFont(scint_font)->pfd) {
+		// Current font is Pango font
+		gtk_widget_modify_font(PWidget(list), PFont(scint_font)->pfd);
 	}
 #endif
 }
@@ -1822,6 +1961,10 @@ void ListBoxX::SetVisibleRows(int rows) {
 	desiredVisibleRows = rows;
 }
 
+int ListBoxX::GetVisibleRows() const {
+	return desiredVisibleRows;
+}
+
 PRectangle ListBoxX::GetDesiredRect() {
 	// Before any size allocated pretend its 100 wide so not scrolled
 	PRectangle rc(0, 0, 100, 100);
@@ -1833,18 +1976,27 @@ PRectangle ListBoxX::GetDesiredRect() {
 		GtkRequisition req;
 		int height;
 
+		// First calculate height of the clist for our desired visible
+		// row count otherwise it tries to expand to the total # of rows
 #if GTK_MAJOR_VERSION < 2
-
 		int ythickness = PWidget(list)->style->klass->ythickness;
-#else
-
-		int ythickness = PWidget(list)->style->ythickness;
-#endif
-		// First calculate height of the clist for our desired visible row count otherwise it tries to expand to the total # of rows
 		height = (rows * GTK_CLIST(list)->row_height
 		          + rows + 1
 		          + 2 * (ythickness
 		                 + GTK_CONTAINER(PWidget(list))->border_width));
+#else
+		// Get cell height
+		int row_width=0;
+		int row_height=0;
+		GtkTreeViewColumn * column =
+			gtk_tree_view_get_column(GTK_TREE_VIEW(list), 0);
+		gtk_tree_view_column_cell_get_size(column, NULL,
+			NULL, NULL, &row_width, &row_height);
+		int ythickness = PWidget(list)->style->ythickness;
+		height = (rows * row_height
+		          + 2 * (ythickness
+		                 + GTK_CONTAINER(PWidget(list))->border_width + 1));
+#endif
 		gtk_widget_set_usize(GTK_WIDGET(PWidget(list)), -1, height);
 
 		// Get the size of the scroller because we set usize on the window
@@ -1868,11 +2020,20 @@ int ListBoxX::CaretFromEdge() {
 }
 
 void ListBoxX::Clear() {
+#if GTK_MAJOR_VERSION < 2
 	gtk_clist_clear(GTK_CLIST(list));
+#else
+	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(list));
+	gtk_list_store_clear(GTK_LIST_STORE(model));
+#endif
 	maxItemCharacters = 0;
 }
 
+#if GTK_MAJOR_VERSION < 2
 static void init_pixmap(ListImage *list_image, GtkWidget *window) {
+#else
+static void init_pixmap(ListImage *list_image) {
+#endif
 	const char *textForm = list_image->xpm_data;
 	const char * const * xpm_lineform = reinterpret_cast<const char * const *>(textForm);
 	const char **xpm_lineformfromtext = 0;
@@ -1887,6 +2048,7 @@ static void init_pixmap(ListImage *list_image, GtkWidget *window) {
 	}
 
 	// Drop any existing pixmap/bitmap as data may have changed
+#if GTK_MAJOR_VERSION < 2
 	if (list_image->pixmap)
 		gdk_pixmap_unref(list_image->pixmap);
 	list_image->pixmap = NULL;
@@ -1902,18 +2064,25 @@ static void init_pixmap(ListImage *list_image, GtkWidget *window) {
 			gdk_bitmap_unref(list_image->bitmap);
 		list_image->bitmap = NULL;
 	}
+#else
+	if (list_image->pixbuf)
+		gdk_pixbuf_unref(list_image->pixbuf);
+	list_image->pixbuf =
+		gdk_pixbuf_new_from_xpm_data((const gchar**)xpm_lineform);
+#endif
 	delete []xpm_lineformfromtext;
 }
 
 #define SPACING 5
 
 void ListBoxX::Append(char *s, int type) {
-	char * szs[] = { s, NULL };
 	ListImage *list_image = NULL;
 	if ((type >= 0) && pixhash) {
 		list_image = (ListImage *) g_hash_table_lookup((GHashTable *) pixhash
 		             , (gconstpointer) GINT_TO_POINTER(type));
 	}
+#if GTK_MAJOR_VERSION < 2
+	char * szs[] = { s, NULL };
 	int rownum = gtk_clist_append(GTK_CLIST(list), szs);
 	if (list_image) {
 		if (NULL == list_image->pixmap)
@@ -1921,6 +2090,27 @@ void ListBoxX::Append(char *s, int type) {
 		gtk_clist_set_pixtext(GTK_CLIST(list), rownum, 0, s, SPACING
 		                      , list_image->pixmap, list_image->bitmap);
 	}
+#else
+	GtkTreeIter iter;
+	GtkListStore *store =
+		GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(list)));
+	gtk_list_store_append(GTK_LIST_STORE(store), &iter);
+	if (list_image) {
+		if (NULL == list_image->pixbuf)
+			init_pixmap(list_image);
+		if (list_image->pixbuf) {
+			gtk_list_store_set(GTK_LIST_STORE(store), &iter,
+								PIXBUF_COLUMN, list_image->pixbuf,
+								TEXT_COLUMN, s, -1);
+		} else {
+			gtk_list_store_set(GTK_LIST_STORE(store), &iter,
+								TEXT_COLUMN, s, -1);
+		}
+	} else {
+			gtk_list_store_set(GTK_LIST_STORE(store), &iter,
+								TEXT_COLUMN, s, -1);
+	}
+#endif
 	size_t len = strlen(s);
 	if (maxItemCharacters < len)
 		maxItemCharacters = len;
@@ -1928,20 +2118,95 @@ void ListBoxX::Append(char *s, int type) {
 
 int ListBoxX::Length() {
 	if (id)
+#if GTK_MAJOR_VERSION < 2
 		return GTK_CLIST(list)->rows;
+#else
+		return gtk_tree_model_iter_n_children(gtk_tree_view_get_model
+											   (GTK_TREE_VIEW(list)), NULL);
+#endif
 	return 0;
 }
 
 void ListBoxX::Select(int n) {
-	gtk_clist_select_row(GTK_CLIST(list), n, 0);
-	gtk_clist_moveto(GTK_CLIST(list), n, 0, 0.5, 0.5);
+#if GTK_MAJOR_VERSION < 2
+	if (n == -1) {
+		gtk_clist_unselect_row(GTK_CLIST(list), current, 0);
+	} else {
+		gtk_clist_select_row(GTK_CLIST(list), n, 0);
+		gtk_clist_moveto(GTK_CLIST(list), n, 0, 0.5, 0.5);
+	}
+#else
+	GtkTreeIter iter;
+	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(list));
+	GtkTreeSelection *selection =
+		gtk_tree_view_get_selection(GTK_TREE_VIEW(list));
+
+	if (n < 0) {
+		gtk_tree_selection_unselect_all(selection);
+		return;
+	}
+
+	bool valid = gtk_tree_model_iter_nth_child(model, &iter, NULL, n);
+	if (valid) {
+		gtk_tree_selection_select_iter(selection, &iter);
+
+		// Move the scrollbar to show the selection.
+		int total = Length();
+		GtkAdjustment *adj =
+			gtk_tree_view_get_vadjustment(GTK_TREE_VIEW(list));
+		gfloat value = ((gfloat)n / total) * (adj->upper - adj->lower)
+							+ adj->lower - adj->page_size / 2;
+
+		// Get cell height
+		int row_width;
+		int row_height;
+		GtkTreeViewColumn * column =
+			gtk_tree_view_get_column(GTK_TREE_VIEW(list), 0);
+		gtk_tree_view_column_cell_get_size(column, NULL, NULL,
+											NULL, &row_width, &row_height);
+
+		int rows = Length();
+		if ((rows == 0) || (rows > desiredVisibleRows))
+			rows = desiredVisibleRows;
+		if (rows & 0x1) {
+			// Odd rows to display -- We are now in the middle.
+			// Align it so that we don't chop off rows.
+			value += (gfloat)row_height / 2.0;
+		}
+		// Clamp it.
+		value = (value < 0)? 0 : value;
+		value = (value > (adj->upper - adj->page_size))?
+					(adj->upper - adj->page_size) : value;
+
+		// Set it.
+		gtk_adjustment_set_value(adj, value);
+	} else {
+		gtk_tree_selection_unselect_all(selection);
+	}
+#endif
 }
 
 int ListBoxX::GetSelection() {
+#if GTK_MAJOR_VERSION < 2
 	return current;
+#else
+	GtkTreeIter iter;
+	GtkTreeModel *model;
+	GtkTreeSelection *selection;
+	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(list));
+	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
+		GtkTreePath *path = gtk_tree_model_get_path(model, &iter);
+		int *indices = gtk_tree_path_get_indices(path);
+		// Don't free indices.
+		if (indices)
+			return indices[0];
+	}
+	return -1;
+#endif
 }
 
 int ListBoxX::Find(const char *prefix) {
+#if GTK_MAJOR_VERSION < 2
 	int count = Length();
 	for (int i = 0; i < count; i++) {
 		char *s = 0;
@@ -1950,11 +2215,28 @@ int ListBoxX::Find(const char *prefix) {
 			return i;
 		}
 	}
-	return - 1;
+#else
+	GtkTreeIter iter;
+	GtkTreeModel *model =
+		gtk_tree_view_get_model(GTK_TREE_VIEW(list));
+	bool valid = gtk_tree_model_get_iter_first(model, &iter);
+	int i = 0;
+	while(valid) {
+		gchar *s;
+		gtk_tree_model_get(model, &iter, TEXT_COLUMN, &s, -1);
+		if (s && (0 == strncmp(prefix, s, strlen(prefix)))) {
+			return i;
+		}
+		valid = gtk_tree_model_iter_next(model, &iter);
+		i++;
+	}
+#endif
+	return -1;
 }
 
 void ListBoxX::GetValue(int n, char *value, int len) {
 	char *text = NULL;
+#if GTK_MAJOR_VERSION < 2
 	GtkCellType type = gtk_clist_get_cell_type(GTK_CLIST(list), n, 0);
 	switch (type) {
 	case GTK_CELL_TEXT:
@@ -1966,6 +2248,14 @@ void ListBoxX::GetValue(int n, char *value, int len) {
 	default:
 		break;
 	}
+#else
+	GtkTreeIter iter;
+	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(list));
+	bool valid = gtk_tree_model_iter_nth_child(model, &iter, NULL, n);
+	if (valid) {
+		gtk_tree_model_get(model, &iter, TEXT_COLUMN, &text, -1);
+	}
+#endif
 	if (text && len > 0) {
 		strncpy(value, text, len);
 		value[len - 1] = '\0';
@@ -1974,10 +2264,6 @@ void ListBoxX::GetValue(int n, char *value, int len) {
 	}
 }
 
-void ListBoxX::Sort() {
-	gtk_clist_sort(GTK_CLIST(list));
-}
-
 // g_return_if_fail causes unnecessary compiler warning in release compile.
 #ifdef _MSC_VER
 #pragma warning(disable: 4127)
@@ -1998,12 +2284,18 @@ void ListBoxX::RegisterImage(int type, const char *xpm_data) {
 		(gconstpointer) GINT_TO_POINTER(type));
 	if (list_image) {
 		// Drop icon already registered
+#if GTK_MAJOR_VERSION < 2
 		if (list_image->pixmap)
 			gdk_pixmap_unref(list_image->pixmap);
 		list_image->pixmap = 0;
 		if (list_image->bitmap)
 			gdk_bitmap_unref(list_image->bitmap);
 		list_image->bitmap = 0;
+#else
+		if (list_image->pixbuf)
+			gdk_pixbuf_unref(list_image->pixbuf);
+		list_image->pixbuf = NULL;
+#endif
 		list_image->xpm_data = xpm_data;
 	} else {
 		list_image = g_new0(ListImage, 1);
@@ -2017,6 +2309,36 @@ void ListBoxX::ClearRegisteredImages() {
 	xset.Clear();
 }
 
+void ListBoxX::SetList(const char* list, char separator, char typesep) {
+	Clear();
+	int count = strlen(list) + 1;
+	char *words = new char[count];
+	if (words) {
+		memcpy(words, list, count);
+		char *startword = words;
+		char *numword = NULL;
+		int i = 0;
+		for (; words[i]; i++) {
+			if (words[i] == separator) {
+				words[i] = '\0';
+				if (numword)
+					*numword = '\0';
+				Append(startword, numword?atoi(numword + 1):-1);
+				startword = words + i + 1;
+				numword = NULL;
+			} else if (words[i] == typesep) {
+				numword = words + i;
+			}
+		}
+		if (startword) {
+			if (numword)
+				*numword = '\0';
+			Append(startword, numword?atoi(numword + 1):-1);
+		}
+		delete []words;
+	}
+}
+
 Menu::Menu() : id(0) {}
 
 void Menu::CreatePopUp() {
@@ -2044,7 +2366,12 @@ void Menu::Show(Point pt, Window &) {
 	if ((pt.y + requisition.height) > screenHeight) {
 		pt.y = screenHeight - requisition.height;
 	}
-	gtk_item_factory_popup(factory, pt.x - 4, pt.y, 3, 0);
+#if GTK_MAJOR_VERSION >= 2
+	gtk_item_factory_popup(factory, pt.x - 4, pt.y - 4, 3, 
+		gtk_get_current_event_time());
+#else
+	gtk_item_factory_popup(factory, pt.x - 4, pt.y - 4, 3, 0);
+#endif
 }
 
 ElapsedTime::ElapsedTime() {
@@ -2158,20 +2485,40 @@ long Platform::SendScintillaPointer(
 	                              reinterpret_cast<sptr_t>(lParam));
 }
 
-bool Platform::IsDBCSLeadByte(int /*codePage*/, char /*ch*/) {
+bool Platform::IsDBCSLeadByte(int /* codePage */, char /* ch */) {
 	return false;
 }
 
-int Platform::DBCSCharLength(int /*codePage*/, const char *s) {
+#if GTK_MAJOR_VERSION < 2
+int Platform::DBCSCharLength(int, const char *s) {
 	int bytes = mblen(s, MB_CUR_MAX);
 	if (bytes >= 1)
 		return bytes;
 	else
 		return 1;
 }
+#else
+int Platform::DBCSCharLength(int codePage, const char *s) {
+	if (codePage == 999932) {
+		// Experimental and disabled code - change 999932 to 932 above to
+		// enable locale avoiding but expensive character length determination.
+		// Avoid locale with explicit use of iconv
+		Converter convMeasure("UCS-2", CharacterSetID(SC_CHARSET_SHIFTJIS));
+		size_t lenChar = MultiByteLenFromIconv(convMeasure, s, strlen(s));
+		return lenChar;
+	} else {
+		int bytes = mblen(s, MB_CUR_MAX);
+		if (bytes >= 1)
+			return bytes;
+		else
+			return 1;
+	}
+}
+#endif
 
 int Platform::DBCSCharMaxLength() {
 	return MB_CUR_MAX;
+	//return 2;
 }
 
 // These are utility functions not really tied to a platform
