@@ -1,5 +1,5 @@
 /*
-	Copyright (C) 2003  Sridhar <r_sridhar@users.sf.net>
+	Copyright (C) 2003  Sridhar Ratna <srid@nearfar.org>
 	
 	This file is part of GNUSim8085.
 
@@ -54,7 +54,10 @@ get_hex_list (const gchar * str, gint * total)
 		(*total)++;
 		
 		/* add hex */
-		g_string_append_printf (r, " %02X", val);
+		if (val > 0x0f)
+			g_string_append_printf (r, " %02X", val);
+		else
+			g_string_append_printf (r, " 0%X", val);
 		p++;
 	}
 	
@@ -102,14 +105,38 @@ asm_listing_get_binary_array (AsmSource *src, GString **listing)
 			IdOpcode *ido = entry->s_op_id_opcode;
 			g_assert (ido);
 			
-			g_string_append_printf (listing[pto], "%4X\t%2X", entry->address, entry->s_op);
+			if ( entry->s_op > 0x0f )
+			{
+				g_string_append_printf (listing[pto], "%4X\t%2X", entry->address, entry->s_op);
+			}
+			else
+			{
+				g_string_append_printf (listing[pto], "%4X\t0%X", entry->address, entry->s_op);
+			}
+ 			
 			
 			/* if args */
 			if ( ido->user_args )
 			{
-				g_string_append_printf(listing[pto], " %2X", entry->b_opnd1);
+				if ( entry->b_opnd1 > 0x0f )
+				{
+					g_string_append_printf(listing[pto], " %2X", entry->b_opnd1);
+				}
+				else
+				{
+					g_string_append_printf(listing[pto], " 0%X", entry->b_opnd1);
+				}
 				if ( ido->user_args == 2 )
-					g_string_append_printf(listing[pto], " %2X", entry->b_opnd2);
+				{
+					if ( entry->b_opnd1 > 0x0f )
+					{
+						g_string_append_printf(listing[pto], " %2X", entry->b_opnd2);
+					}
+					else
+					{
+						g_string_append_printf(listing[pto], " 0%X", entry->b_opnd2);
+					}
+				}
 			}
 		}
 		else /* if pseudo */
